import asyncio
import logging
import os.path
import sys
from typing import BinaryIO
from typing import Optional

from mitmproxy import command
from mitmproxy import ctx
from mitmproxy import exceptions
from mitmproxy import flowfilter
from mitmproxy import io
from mitmproxy.utils import asyncio_utils

logger = logging.getLogger(__name__)


class ReadFile:
    """
    An addon that handles reading from file on startup.
    """

    def __init__(self):
        self.filter = None
        self._read_task: asyncio.Task | None = None

    def load(self, loader):
        loader.add_option("rfile", Optional[str], None, "Read flows from file.")
        loader.add_option(
            "readfile_filter", Optional[str], None, "Read only matching flows."
        )

    def configure(self, updated):
        if "readfile_filter" in updated:
            if ctx.options.readfile_filter:
                try:
                    self.filter = flowfilter.parse(ctx.options.readfile_filter)
                except ValueError as e:
                    raise exceptions.OptionsError(str(e)) from e
            else:
                self.filter = None

    async def load_flows(self, fo: BinaryIO) -> int:
        cnt = 0
        freader = io.FlowReader(fo)
        try:
            for flow in freader.stream():
                if self.filter and not self.filter(flow):
                    continue
                await ctx.master.load_flow(flow)
                cnt += 1
        except (OSError, exceptions.FlowReadException) as e:
            if cnt:
                logging.warning("Flow file corrupted - loaded %i flows." % cnt)
            else:
                logging.error("Flow file corrupted.")
            raise exceptions.FlowReadException(str(e)) from e
        else:
            return cnt

    async def load_flows_from_path(self, path: str) -> int:
        path = os.path.expanduser(path)
        try:
            with open(path, "rb") as f:
                return await self.load_flows(f)
        except OSError as e:
            logging.error(f"Cannot load flows: {e}")
            raise exceptions.FlowReadException(str(e)) from e

    async def doread(self, rfile: str) -> None:
        try:
            await self.load_flows_from_path(rfile)
        except exceptions.FlowReadException as e:
            logger.exception(f"Failed to read {ctx.options.rfile}: {e}")

    def running(self):
        if ctx.options.rfile:
            self._read_task = asyncio_utils.create_task(
                self.doread(ctx.options.rfile),
                name="readfile",
                keep_ref=False,
            )

    @command.command("readfile.reading")
    def reading(self) -> bool:
        return bool(self._read_task and not self._read_task.done())


class ReadFileStdin(ReadFile):
    """Support the special case of "-" for reading from stdin"""

    async def load_flows_from_path(self, path: str) -> int:
        if path == "-":  # pragma: no cover
            # Need to think about how to test this. This function is scheduled
            # onto the event loop, where a sys.stdin mock has no effect.
            return await self.load_flows(sys.stdin.buffer)
        else:
            return await super().load_flows_from_path(path)



"""
This module provides a method to detect if a given file object supports virtual terminal escape codes.
"""

import os
import sys
from typing import IO

if os.name == "nt":
    from ctypes import byref  # type: ignore
    from ctypes import windll  # type: ignore
    from ctypes.wintypes import BOOL
    from ctypes.wintypes import DWORD
    from ctypes.wintypes import HANDLE
    from ctypes.wintypes import LPDWORD

    ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12

    # https://docs.microsoft.com/de-de/windows/console/getstdhandle
    GetStdHandle = windll.kernel32.GetStdHandle
    GetStdHandle.argtypes = [DWORD]
    GetStdHandle.restype = HANDLE

    # https://docs.microsoft.com/de-de/windows/console/getconsolemode
    GetConsoleMode = windll.kernel32.GetConsoleMode
    GetConsoleMode.argtypes = [HANDLE, LPDWORD]
    GetConsoleMode.restype = BOOL

    # https://docs.microsoft.com/de-de/windows/console/setconsolemode
    SetConsoleMode = windll.kernel32.SetConsoleMode
    SetConsoleMode.argtypes = [HANDLE, DWORD]
    SetConsoleMode.restype = BOOL

    def ensure_supported(f: IO[str]) -> bool:
        if not f.isatty():
            return False
        if f == sys.stdout:
            h = STD_OUTPUT_HANDLE
        elif f == sys.stderr:
            h = STD_ERROR_HANDLE
        else:
            return False

        handle = GetStdHandle(h)
        console_mode = DWORD()
        ok = GetConsoleMode(handle, byref(console_mode))
        if not ok:
            return False

        ok = SetConsoleMode(
            handle, console_mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING
        )
        return ok

else:

    def ensure_supported(f: IO[str]) -> bool:
        return f.isatty()



from ._api import Contentview
from ._api import Metadata


class RawContentview(Contentview):
    def prettify(self, data: bytes, metadata: Metadata) -> str:
        return data.decode("utf-8", "backslashreplace")

    def render_priority(
        self,
        data: bytes,
        metadata: Metadata,
    ) -> float:
        return 0.1


raw = RawContentview()



# Covered status codes:
# - official HTTP status codes: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
# - custom codes:
#   - 444 No Response
#   - 499 Client Closed Request
CONTINUE = 100
SWITCHING = 101
PROCESSING = 102
EARLY_HINTS = 103

OK = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206
MULTI_STATUS = 207
ALREADY_REPORTED = 208
IM_USED = 226

MULTIPLE_CHOICE = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
USE_PROXY = 305
TEMPORARY_REDIRECT = 307
PERMANENT_REDIRECT = 308

BAD_REQUEST = 400
UNAUTHORIZED = 401
PAYMENT_REQUIRED = 402
FORBIDDEN = 403
NOT_FOUND = 404
NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTH_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409
GONE = 410
LENGTH_REQUIRED = 411
PRECONDITION_FAILED = 412
PAYLOAD_TOO_LARGE = 413
REQUEST_URI_TOO_LONG = 414
UNSUPPORTED_MEDIA_TYPE = 415
REQUESTED_RANGE_NOT_SATISFIABLE = 416
EXPECTATION_FAILED = 417
IM_A_TEAPOT = 418
MISDIRECTED_REQUEST = 421
UNPROCESSABLE_CONTENT = 422
LOCKED = 423
FAILED_DEPENDENCY = 424
TOO_EARLY = 425
UPGRADE_REQUIRED = 426
PRECONDITION_REQUIRED = 428
TOO_MANY_REQUESTS = 429
REQUEST_HEADER_FIELDS_TOO_LARGE = 431
UNAVAILABLE_FOR_LEGAL_REASONS = 451
NO_RESPONSE = 444
CLIENT_CLOSED_REQUEST = 499

INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501
BAD_GATEWAY = 502
SERVICE_UNAVAILABLE = 503
GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
VARIANT_ALSO_NEGOTIATES = 506
INSUFFICIENT_STORAGE_SPACE = 507
LOOP_DETECTED = 508
NOT_EXTENDED = 510
NETWORK_AUTHENTICATION_REQUIRED = 511

RESPONSES = {
    # 100
    CONTINUE: "Continue",
    SWITCHING: "Switching Protocols",
    PROCESSING: "Processing",
    EARLY_HINTS: "Early Hints",
    # 200
    OK: "OK",
    CREATED: "Created",
    ACCEPTED: "Accepted",
    NON_AUTHORITATIVE_INFORMATION: "Non-Authoritative Information",
    NO_CONTENT: "No Content",
    RESET_CONTENT: "Reset Content",
    PARTIAL_CONTENT: "Partial Content",
    MULTI_STATUS: "Multi-Status",
    ALREADY_REPORTED: "Already Reported",
    IM_USED: "IM Used",
    # 300
    MULTIPLE_CHOICE: "Multiple Choices",
    MOVED_PERMANENTLY: "Moved Permanently",
    FOUND: "Found",
    SEE_OTHER: "See Other",
    NOT_MODIFIED: "Not Modified",
    USE_PROXY: "Use Proxy",
    # 306 not defined??
    TEMPORARY_REDIRECT: "Temporary Redirect",
    PERMANENT_REDIRECT: "Permanent Redirect",
    # 400
    BAD_REQUEST: "Bad Request",
    UNAUTHORIZED: "Unauthorized",
    PAYMENT_REQUIRED: "Payment Required",
    FORBIDDEN: "Forbidden",
    NOT_FOUND: "Not Found",
    NOT_ALLOWED: "Method Not Allowed",
    NOT_ACCEPTABLE: "Not Acceptable",
    PROXY_AUTH_REQUIRED: "Proxy Authentication Required",
    REQUEST_TIMEOUT: "Request Time-out",
    CONFLICT: "Conflict",
    GONE: "Gone",
    LENGTH_REQUIRED: "Length Required",
    PRECONDITION_FAILED: "Precondition Failed",
    PAYLOAD_TOO_LARGE: "Payload Too Large",
    REQUEST_URI_TOO_LONG: "Request-URI Too Long",
    UNSUPPORTED_MEDIA_TYPE: "Unsupported Media Type",
    REQUESTED_RANGE_NOT_SATISFIABLE: "Requested Range not satisfiable",
    EXPECTATION_FAILED: "Expectation Failed",
    IM_A_TEAPOT: "I'm a teapot",
    MISDIRECTED_REQUEST: "Misdirected Request",
    UNPROCESSABLE_CONTENT: "Unprocessable Content",
    LOCKED: "Locked",
    FAILED_DEPENDENCY: "Failed Dependency",
    TOO_EARLY: "Too Early",
    UPGRADE_REQUIRED: "Upgrade Required",
    PRECONDITION_REQUIRED: "Precondition Required",
    TOO_MANY_REQUESTS: "Too Many Requests",
    REQUEST_HEADER_FIELDS_TOO_LARGE: "Request Header Fields Too Large",
    UNAVAILABLE_FOR_LEGAL_REASONS: "Unavailable For Legal Reasons",
    NO_RESPONSE: "No Response",
    CLIENT_CLOSED_REQUEST: "Client Closed Request",
    # 500
    INTERNAL_SERVER_ERROR: "Internal Server Error",
    NOT_IMPLEMENTED: "Not Implemented",
    BAD_GATEWAY: "Bad Gateway",
    SERVICE_UNAVAILABLE: "Service Unavailable",
    GATEWAY_TIMEOUT: "Gateway Time-out",
    HTTP_VERSION_NOT_SUPPORTED: "HTTP Version not supported",
    VARIANT_ALSO_NEGOTIATES: "Variant Also Negotiates",
    INSUFFICIENT_STORAGE_SPACE: "Insufficient Storage Space",
    LOOP_DETECTED: "Loop Detected",
    NOT_EXTENDED: "Not Extended",
    NETWORK_AUTHENTICATION_REQUIRED: "Network Authentication Required",
}



A = 1
NS = 2
MD = 3
MF = 4
CNAME = 5
SOA = 6
MB = 7
MG = 8
MR = 9
NULL = 10
WKS = 11
PTR = 12
HINFO = 13
MINFO = 14
MX = 15
TXT = 16
RP = 17
AFSDB = 18
X25 = 19
ISDN = 20
RT = 21
NSAP = 22
NSAP_PTR = 23
SIG = 24
KEY = 25
PX = 26
GPOS = 27
AAAA = 28
LOC = 29
NXT = 30
EID = 31
NIMLOC = 32
SRV = 33
ATMA = 34
NAPTR = 35
KX = 36
CERT = 37
A6 = 38
DNAME = 39
SINK = 40
OPT = 41
APL = 42
DS = 43
SSHFP = 44
IPSECKEY = 45
RRSIG = 46
NSEC = 47
DNSKEY = 48
DHCID = 49
NSEC3 = 50
NSEC3PARAM = 51
TLSA = 52
SMIMEA = 53
HIP = 55
NINFO = 56
RKEY = 57
TALINK = 58
CDS = 59
CDNSKEY = 60
OPENPGPKEY = 61
CSYNC = 62
ZONEMD = 63
SVCB = 64
HTTPS = 65
SPF = 99
UINFO = 100
UID = 101
GID = 102
UNSPEC = 103
NID = 104
L32 = 105
L64 = 106
LP = 107
EUI48 = 108
EUI64 = 109
TKEY = 249
TSIG = 250
IXFR = 251
AXFR = 252
MAILB = 253
MAILA = 254
ANY = 255
URI = 256
CAA = 257
AVC = 258
DOA = 259
AMTRELAY = 260
TA = 32768
DLV = 32769

_STRINGS = {
    A: "A",
    NS: "NS",
    MD: "MD",
    MF: "MF",
    CNAME: "CNAME",
    SOA: "SOA",
    MB: "MB",
    MG: "MG",
    MR: "MR",
    NULL: "NULL",
    WKS: "WKS",
    PTR: "PTR",
    HINFO: "HINFO",
    MINFO: "MINFO",
    MX: "MX",
    TXT: "TXT",
    RP: "RP",
    AFSDB: "AFSDB",
    X25: "X25",
    ISDN: "ISDN",
    RT: "RT",
    NSAP: "NSAP",
    NSAP_PTR: "NSAP_PTR",
    SIG: "SIG",
    KEY: "KEY",
    PX: "PX",
    GPOS: "GPOS",
    AAAA: "AAAA",
    LOC: "LOC",
    NXT: "NXT",
    EID: "EID",
    NIMLOC: "NIMLOC",
    SRV: "SRV",
    ATMA: "ATMA",
    NAPTR: "NAPTR",
    KX: "KX",
    CERT: "CERT",
    A6: "A6",
    DNAME: "DNAME",
    SINK: "SINK",
    OPT: "OPT",
    APL: "APL",
    DS: "DS",
    SSHFP: "SSHFP",
    IPSECKEY: "IPSECKEY",
    RRSIG: "RRSIG",
    NSEC: "NSEC",
    DNSKEY: "DNSKEY",
    DHCID: "DHCID",
    NSEC3: "NSEC3",
    NSEC3PARAM: "NSEC3PARAM",
    TLSA: "TLSA",
    SMIMEA: "SMIMEA",
    HIP: "HIP",
    NINFO: "NINFO",
    RKEY: "RKEY",
    TALINK: "TALINK",
    CDS: "CDS",
    CDNSKEY: "CDNSKEY",
    OPENPGPKEY: "OPENPGPKEY",
    CSYNC: "CSYNC",
    ZONEMD: "ZONEMD",
    SVCB: "SVCB",
    HTTPS: "HTTPS",
    SPF: "SPF",
    UINFO: "UINFO",
    UID: "UID",
    GID: "GID",
    UNSPEC: "UNSPEC",
    NID: "NID",
    L32: "L32",
    L64: "L64",
    LP: "LP",
    EUI48: "EUI48",
    EUI64: "EUI64",
    TKEY: "TKEY",
    TSIG: "TSIG",
    IXFR: "IXFR",
    AXFR: "AXFR",
    MAILB: "MAILB",
    MAILA: "MAILA",
    ANY: "ANY",
    URI: "URI",
    CAA: "CAA",
    AVC: "AVC",
    DOA: "DOA",
    AMTRELAY: "AMTRELAY",
    TA: "TA",
    DLV: "DLV",
}
_INTS = {v: k for k, v in _STRINGS.items()}


def to_str(type_: int) -> str:
    return _STRINGS.get(type_, f"TYPE({type_})")


def from_str(type_: str) -> int:
    try:
        return _INTS[type_]
    except KeyError:
        return int(type_.removeprefix("TYPE(").removesuffix(")"))



from __future__ import annotations

import pprint
import textwrap
import typing
from collections.abc import Sequence
from typing import Optional

import urwid

from mitmproxy import exceptions
from mitmproxy import optmanager
from mitmproxy.tools.console import layoutwidget
from mitmproxy.tools.console import overlay
from mitmproxy.tools.console import signals

HELP_HEIGHT = 5


def can_edit_inplace(opt):
    if opt.choices:
        return False
    if opt.typespec in [str, int, Optional[str], Optional[int]]:
        return True


class OptionItem(urwid.WidgetWrap):
    def __init__(self, walker, opt, focused, namewidth, editing):
        self.walker, self.opt, self.focused = walker, opt, focused
        self.namewidth = namewidth
        self.editing = editing
        super().__init__(self.get_widget())

    def get_widget(self):
        val = self.opt.current()
        if self.opt.typespec is bool:
            displayval = "true" if val else "false"
        elif not val:
            displayval = ""
        elif self.opt.typespec == Sequence[str]:
            displayval = pprint.pformat(val, indent=1)
        else:
            displayval = str(val)

        changed = self.walker.master.options.has_changed(self.opt.name)
        if self.focused:
            valstyle = "option_active_selected" if changed else "option_selected"
        else:
            valstyle = "option_active" if changed else "text"

        if self.editing:
            valw = urwid.Edit(edit_text=displayval)
        else:
            valw = urwid.AttrMap(
                urwid.Padding(urwid.Text([(valstyle, displayval)])), valstyle
            )

        return urwid.Columns(
            [
                (
                    self.namewidth,
                    urwid.Text([("title", self.opt.name.ljust(self.namewidth))]),
                ),
                valw,
            ],
            dividechars=2,
            focus_column=1,
        )

    def get_edit_text(self):
        return self._w[1].get_edit_text()

    def selectable(self):
        return True

    def keypress(self, size, key):
        if self.editing:
            self._w[1].keypress(size, key)
            return
        return key


class OptionListWalker(urwid.ListWalker):
    def __init__(self, master, help_widget: OptionHelp):
        self.master = master
        self.help_widget = help_widget

        self.index = 0
        self.focusobj = None

        self.opts = sorted(master.options.keys())
        self.maxlen = max(len(i) for i in self.opts)
        self.editing = False
        self.set_focus(0)
        self.master.options.changed.connect(self.sig_mod)

    def sig_mod(self, *args, **kwargs):
        self.opts = sorted(self.master.options.keys())
        self.maxlen = max(len(i) for i in self.opts)
        self._modified()
        self.set_focus(self.index)

    def start_editing(self):
        self.editing = True
        self.focus_obj = self._get(self.index, True)
        self._modified()

    def stop_editing(self):
        self.editing = False
        self.focus_obj = self._get(self.index, False)
        self.set_focus(self.index)
        self._modified()

    def get_edit_text(self):
        return self.focus_obj.get_edit_text()

    def _get(self, pos, editing):
        name = self.opts[pos]
        opt = self.master.options._options[name]
        return OptionItem(self, opt, pos == self.index, self.maxlen, editing)

    def get_focus(self):
        return self.focus_obj, self.index

    def set_focus(self, index):
        self.editing = False
        name = self.opts[index]
        opt = self.master.options._options[name]
        self.index = index
        self.focus_obj = self._get(self.index, self.editing)
        self.help_widget.update_help_text(opt.help)
        self._modified()

    def get_next(self, pos):
        if pos >= len(self.opts) - 1:
            return None, None
        pos = pos + 1
        return self._get(pos, False), pos

    def get_prev(self, pos):
        pos = pos - 1
        if pos < 0:
            return None, None
        return self._get(pos, False), pos

    def positions(self, reverse=False):
        if reverse:
            return reversed(range(len(self.opts)))
        else:
            return range(len(self.opts))


class OptionsList(urwid.ListBox):
    def __init__(self, master, help_widget: OptionHelp):
        self.master = master
        self.walker = OptionListWalker(master, help_widget)
        super().__init__(self.walker)

    def save_config(self, path):
        try:
            optmanager.save(self.master.options, path)
        except exceptions.OptionsError as e:
            signals.status_message.send(message=str(e))

    def keypress(self, size, key):
        if self.walker.editing:
            if key == "enter":
                foc, idx = self.get_focus()
                v = self.walker.get_edit_text()
                try:
                    self.master.options.set(f"{foc.opt.name}={v}")
                except exceptions.OptionsError as v:
                    signals.status_message.send(message=str(v))
                self.walker.stop_editing()
                return None
            elif key == "esc":
                self.walker.stop_editing()
                return None
        else:
            if key == "m_start":
                self.set_focus(0)
                self.walker._modified()
            elif key == "m_end":
                self.set_focus(len(self.walker.opts) - 1)
                self.walker._modified()
            elif key == "m_select":
                foc, idx = self.get_focus()
                if foc.opt.typespec is bool:
                    self.master.options.toggler(foc.opt.name)()
                    # Bust the focus widget cache
                    self.set_focus(self.walker.index)
                elif can_edit_inplace(foc.opt):
                    self.walker.start_editing()
                    self.walker._modified()
                elif foc.opt.choices:
                    self.master.overlay(
                        overlay.Chooser(
                            self.master,
                            foc.opt.name,
                            foc.opt.choices,
                            foc.opt.current(),
                            self.master.options.setter(foc.opt.name),
                        )
                    )
                elif foc.opt.typespec in (Sequence[str], typing.Sequence[str]):
                    self.master.overlay(
                        overlay.OptionsOverlay(
                            self.master,
                            foc.opt.name,
                            foc.opt.current(),
                            HELP_HEIGHT + 5,
                        ),
                        valign="top",
                    )
                else:
                    raise NotImplementedError()
        return super().keypress(size, key)


class OptionHelp(urwid.Frame):
    def __init__(self, master):
        self.master = master
        super().__init__(self.widget(""))
        self.set_active(False)

    def set_active(self, val):
        h = urwid.Text("Option Help")
        style = "heading" if val else "heading_inactive"
        self.header = urwid.AttrMap(h, style)

    def widget(self, txt):
        cols, _ = self.master.ui.get_cols_rows()
        return urwid.ListBox([urwid.Text(i) for i in textwrap.wrap(txt, cols)])

    def update_help_text(self, txt: str) -> None:
        self.body = self.widget(txt)


class Options(urwid.Pile, layoutwidget.LayoutWidget):
    title = "Options"
    keyctx = "options"

    focus_position: int

    def __init__(self, master):
        oh = OptionHelp(master)
        self.optionslist = OptionsList(master, oh)
        super().__init__(
            [
                self.optionslist,
                (HELP_HEIGHT, oh),
            ]
        )
        self.master = master

    def current_name(self):
        foc, idx = self.optionslist.get_focus()
        return foc.opt.name

    def keypress(self, size, key):
        if key == "m_next":
            self.focus_position = (self.focus_position + 1) % len(self.widget_list)
            self.widget_list[1].set_active(self.focus_position == 1)
            key = None

        # This is essentially a copypasta from urwid.Pile's keypress handler.
        # So much for "closed for modification, but open for extension".
        item_rows = None
        if len(size) == 2:
            item_rows = self.get_item_rows(size, focus=True)
        tsize = self.get_item_size(size, self.focus_position, True, item_rows)
        return self.focus.keypress(tsize, key)



"""
This module provides a @concurrent decorator primitive to
offload computations from mitmproxy's main master thread.
"""

import asyncio
import inspect

from mitmproxy import hooks


def concurrent(fn):
    if fn.__name__ not in set(hooks.all_hooks.keys()) - {"load", "configure"}:
        raise NotImplementedError(
            "Concurrent decorator not supported for '%s' method." % fn.__name__
        )

    async def _concurrent(*args):
        def run():
            if inspect.iscoroutinefunction(fn):
                # Run the async function in a new event loop
                loop = asyncio.new_event_loop()
                try:
                    loop.run_until_complete(fn(*args))
                finally:
                    loop.close()
            else:
                fn(*args)

        await asyncio.get_running_loop().run_in_executor(None, run)

    return _concurrent



from __future__ import annotations

import logging
import typing
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

from mitmproxy import http
from mitmproxy import tcp
from mitmproxy import udp
from mitmproxy.dns import DNSMessage
from mitmproxy.flow import Flow
from mitmproxy.websocket import WebSocketMessage

logger = logging.getLogger(__name__)


type SyntaxHighlight = Literal["css", "javascript", "xml", "yaml", "none", "error"]


@typing.runtime_checkable
class Contentview(typing.Protocol):
    """
    Base class for all contentviews.
    """

    @property
    def name(self) -> str:
        """
        The name of this contentview, e.g. "XML/HTML".
        Inferred from the class name by default.
        """
        return type(self).__name__.removesuffix("Contentview")

    @property
    def syntax_highlight(self) -> SyntaxHighlight:
        """Optional syntax highlighting that should be applied to the prettified output."""
        return "none"

    @abstractmethod
    def prettify(
        self,
        data: bytes,
        metadata: Metadata,
    ) -> str:
        """
        Transform raw data into human-readable output.
        May raise an exception (e.g. `ValueError`) if data cannot be prettified.
        """

    def render_priority(
        self,
        data: bytes,
        metadata: Metadata,
    ) -> float:
        """
        Return the priority of this view for rendering `data`.
        If no particular view is chosen by the user, the view with the highest priority is selected.
        If this view does not support the given data, return a float < 0.
        """
        return 0

    def __lt__(self, other):
        return self.name.__lt__(other.name)


@typing.runtime_checkable
class InteractiveContentview(Contentview, typing.Protocol):
    """A contentview that prettifies raw data and allows for interactive editing."""

    @abstractmethod
    def reencode(
        self,
        prettified: str,
        metadata: Metadata,
    ) -> bytes:
        """
        Reencode the given (modified) `prettified` output into the original data format.
        May raise an exception (e.g. `ValueError`) if reencoding failed.
        """


@dataclass
class Metadata:
    """
    Metadata about the data that is being prettified.

    Do not rely on any given attribute to be present.
    """

    flow: Flow | None = None
    """The flow that the data belongs to, if any."""

    content_type: str | None = None
    """The HTTP content type of the data, if any."""
    http_message: http.Message | None = None
    """The HTTP message that the data belongs to, if any."""
    tcp_message: tcp.TCPMessage | None = None
    """The TCP message that the data belongs to, if any."""
    udp_message: udp.UDPMessage | None = None
    """The UDP message that the data belongs to, if any."""
    websocket_message: WebSocketMessage | None = None
    """The websocket message that the data belongs to, if any."""
    dns_message: DNSMessage | None = None
    """The DNS message that the data belongs to, if any."""

    protobuf_definitions: Path | None = None
    """Path to a .proto file that's used to resolve Protobuf field names."""

    original_data: bytes | None = None
    """When reencoding: The original data that was prettified."""


Metadata.__init__.__doc__ = "@private"
